"""
ChatGPT wrapper
"""

from __future__ import annotations
import httpx
import ujson as json
from typing import Any, Self, AsyncGenerator, Literal, Sequence
from typing import TypeAlias, cast, get_args, overload
from enum import Enum
from pydantic import BaseModel as PyBaseModel, Field, validate_call, PrivateAttr
from pydantic import model_validator, ConfigDict
from pydantic import field_serializer, model_serializer
# from PIL import Image
# from pathlib import Path
# from base64 import b64encode, b64decode
# from io import BytesIO
# from vermils.io import aio
from vermils.gadgets import mimics


DEFAULT_CONFIG = ConfigDict(
    validate_assignment=True,
    populate_by_name=True,
    extra="allow")

DEFAULT_HOST = "generativelanguage.googleapis.com"

SafetyBlockThreshold: TypeAlias = Literal[
    "BLOCK_NONE",
    "BLOCK_ONLY_HIGH",
    "BLOCK_MEDIUM_AND_ABOVE",
    "BLOCK_LOW_AND_ABOVE",
]

SafetyViolationProbability: TypeAlias = Literal[
    "NEGLIGIBLE",
    "LOW",
    "MEDIUM",
    "HIGH",
]

SafetyCategory: TypeAlias = Literal[
    "HARM_CATEGORY_HARASSMENT",
    "HARM_CATEGORY_HATE_SPEECH",
    "HARM_CATEGORY_SEXUALLY_EXPLICIT",
    "HARM_CATEGORY_DANGEROUS_CONTENT",
]


@mimics(PyBaseModel.model_dump)
def _dump(m: PyBaseModel, **kw):
    return m.model_dump(exclude_none=True, by_alias=True, mode="json")


class BaseModel(PyBaseModel):
    model_config = DEFAULT_CONFIG


class ModelInfo(BaseModel):
    name: str
    version: str
    display_name: str = Field(alias="displayName")
    description: str
    input_limit: int = Field(alias="inputTokenLimit")
    output_limit: int = Field(alias="outputTokenLimit")
    supported_methods: list[str] = Field(alias="supportedGenerationMethods")
    temperature: float | None = None
    top_p: float | None = Field(alias="topP", default=None)
    top_k: int | None = Field(alias="topK", default=None)


class Role(str, Enum):
    User = "user"
    Model = "model"
    Assistant = "model"
    """Alias for 'Model'"""
    Function = "function"


class Safety(BaseModel):
    category: SafetyCategory
    threshold: SafetyBlockThreshold | None = None
    probability: SafetyViolationProbability | None = Field(exclude=True, default=None)


class Message(BaseModel):

    class BaseSegment(BaseModel):
        model_config = ConfigDict(
            validate_assignment=True,
            populate_by_name=True,
            extra="allow",
            frozen=True,
        )

    class TextSegment(BaseSegment):
        text: str

        def __str__(self):
            return self.text

        def __eq__(self, other):
            if not isinstance(other, Message.TextSegment):
                return NotImplemented
            return self.text == other.text

        def __hash__(self):
            return hash(self.text)

    class ImageSegment(BaseSegment):
        image_url: str
        image_format: str | None = None

        @model_serializer(when_used="json")
        def jsonable_serialize(self):
            return {
                "inlineData": {
                    "mimeType": f"image/{self.image_format or '*'}",
                    "data": self.image_url
                }
            }

        def __str__(self):
            return f"#image({self.image_url})"

        def __eq__(self, other):
            if not isinstance(other, Message.ImageSegment):
                return NotImplemented
            return self.image_url == other.image_url

        def __hash__(self):
            return hash(self.image_url)

    class FuncCallSegment(BaseSegment):
        ...

    class FuncReturnSegment(BaseSegment):
        ...

    role: Role
    content: str | list[TextSegment | ImageSegment
                        | FuncCallSegment | FuncReturnSegment] = Field(alias="parts", default='')
    _tokens: int | None = None
    _hash: int | None = None

    @field_serializer("content", when_used="json")
    def convert_content(self, value, _) -> list[dict]:
        if isinstance(value, str):
            return [{"text": value}]
        return list(_dump(c) for c in value)

    @model_validator(mode="after")
    def post_init(self) -> Self:
        content_hash = (
            hash(self.content)
            if isinstance(self.content, str)
            else hash(tuple(self.content))
        )
        if self._hash is None or self._hash != content_hash:
            self._tokens = None
            self._hash = content_hash

        return self

    def __str__(self) -> str:
        if isinstance(self.content, str):
            return self.content
        return ''.join(str(c) for c in self.content)


SegTypes: TypeAlias = (Message.TextSegment | Message.ImageSegment
                       | Message.FuncCallSegment | Message.FuncReturnSegment)
Prompt: TypeAlias = str | list[SegTypes]


class FullResponse(BaseModel):
    class Candidate(BaseModel):
        content: Message
        finish_reason: str = Field(alias="finishReason")
        index: int
        safety_ratings: list[Safety] = Field(alias="safetyRatings")

    class PromptFeedback(BaseModel):
        safety_ratings: list[Safety] = Field(alias="safetyRatings")

    candidates: list[Candidate]
    prompt_feedback: PromptFeedback = Field(alias="promptFeedback")


async def _better_raise_status(r: httpx.Response):
    if r.status_code != 200:
        await r.aread()
        raise RuntimeError(
            f"{r.status_code} {r.reason_phrase} {r.text}",
        )


class Bot(BaseModel):
    """
    # Gemini bot

    ## Parameters
    - `model` (str): Model to use
    - `api_key` (str): OpenAI API key
    - `prompt` (str): Initial prompt
    - `temperature` (float): The higher the temperature, the crazier the text (0.0 to 1.0)
    - `comp_tokens` (float | int): Reserved tokens for completion, when value is in (0, 1), it represents a ratio,
        [1,-] represents tokens count, 0 for auto mode
    - `top_p` (float): Nucleus sampling: limits the generated guesses to a cumulative probability. (0.0 to 1.0)
    - `proxies` (dict[str, str]): Connection proxies
    - `timeout` (float | None): Connection timeout
    """

    model: str
    api_key: str
    api_host: str = DEFAULT_HOST
    comp_tokens: float = 0.0

    # See https://ai.google.dev/docs/concepts#model_parameters
    max_reply_tokens: int | None = None
    stop: list[str] | None = None
    temperature: float | None = None
    top_p: float | None = None
    top_k: int | None = None
    tools: list[dict[str, Any]] | None = None
    tool_choice: Literal["auto", "none"] | dict[str, Any] = "auto"
    safety_settings: list[Safety] = Field(
        alias="safetySettings", default_factory=lambda: [
            Safety(category=c, threshold="BLOCK_NONE")
            for c in get_args(SafetyCategory)])

    proxy: str | None = None
    timeout: float | None = None
    _cli: httpx.AsyncClient = PrivateAttr(default=None)
    _minfo: ModelInfo | None = None
    _last_proxy: str | None = None
    _last_timeout: float | None = None
    _last_model: str | None = None

    @model_validator(mode="after")
    def post_init(self) -> Self:
        if (cast(None | httpx.AsyncClient, self._cli) is None
                or self.proxy != self._last_proxy or self.timeout != self._last_timeout):
            self.respawn_cli()
        if self.model != self._last_model:
            self._minfo = None
        return self

    def respawn_cli(self, **kw):
        """
        Create a new HTTP client, replacing the old one if it exists.
        """
        self._cli = httpx.AsyncClient(proxy=self.proxy,
                                      timeout=self.timeout,
                                      trust_env=False,
                                      **kw)

    def new_session(self, **kw) -> Session:
        return Session(bot=self, **kw)

    def _get_json(
            self,
            session: Session,
    ) -> dict:
        ret: dict = {
            "contents": list(_dump(m) for m in session.messages),
            "generationConfig": {
                "stopSequences": self.stop,
                "maxOutputTokens": self.max_reply_tokens,
                "topP": self.top_p,
                "topK": self.top_k,
                "temperature": self.temperature,
            }
        }
        if self.safety_settings:
            ret["safetySettings"] = list(_dump(s) for s in self.safety_settings)
        # print(f"json: {ret}")
        return ret

    async def get_model_info(
            self, name: str | None = None, refresh=False) -> ModelInfo:
        name = self.model if name is None else name
        if name == self.model and self._minfo is not None and not refresh:
            return self._minfo
        r = await self._cli.get(
            f"https://{self.api_host}/v1/{self.model}",
            headers={"x-goog-api-key": self.api_key})

        await _better_raise_status(r)

        m = ModelInfo.model_validate_json(r.text)
        if m.name == self.model:
            self._minfo = m
        return m

    async def list_models(self) -> list[ModelInfo]:
        r = await self._cli.get(
            f"https://{self.api_host}/v1/models",
            headers={"x-goog-api-key": self.api_key})

        await _better_raise_status(r)

        ms = list(ModelInfo.model_validate(m) for m in r.json()["models"])
        for m in ms:
            if m.name == self.model:
                self._minfo = m
                break
        return ms

    async def count_tokens(self, msgs: list[Message] | Message) -> int:
        if isinstance(msgs, Message):
            if msgs._tokens is not None:
                return msgs._tokens
            r = await self._cli.post(
                f"https://{self.api_host}/v1/{self.model}:countTokens",
                headers={"x-goog-api-key": self.api_key},
                json={"contents": [_dump(msgs)]})

            await _better_raise_status(r)

            msgs._tokens = cast(int, r.json()["totalTokens"])
            return msgs._tokens

        if all(m._tokens is not None for m in msgs):
            return sum(cast(int, m._tokens) for m in msgs)

        r = await self._cli.post(
            f"https://{self.api_host}/v1/{self.model}:countTokens",
            headers={"x-goog-api-key": self.api_key},
            json={"contents": list(_dump(m) for m in msgs)})

        await _better_raise_status(r)

        return r.json()["totalTokens"]

    async def stream(self,
                     prompt: Prompt,
                     role: Role = Role.User,
                     session: Session | None = None,
                     ) -> AsyncGenerator[str, None]:
        session = self.new_session() if session is None else session
        session.append(Message(role=role, content=prompt))
        await session.trim()

        async with self._cli.stream(
            "POST",
            f"https://{self.api_host}/v1/{self.model}:streamGenerateContent",
            headers={"x-goog-api-key": self.api_key},
            json=self._get_json(session)
        ) as r:
            if r.status_code != 200:
                session.pop()
                await r.aread()
                raise RuntimeError(
                    f"{r.status_code} {r.reason_phrase} {r.text}",
                )
            async for line in r.aiter_lines():
                line = line.strip()
                if line.startswith("\"text\": "):
                    line = json.loads('{'+line+'}')["text"]
                    yield line

    async def send_raw(self,
                       prompt: Prompt,
                       role: Role = Role.User,
                       session: Session | None = None,
                       ) -> FullResponse:
        session = self.new_session() if session is None else session
        session.append(Message(role=role, content=prompt))
        await session.trim()

        r = await self._cli.post(
            f"https://{self.api_host}/v1/{self.model}:generateContent",
            headers={"x-goog-api-key": self.api_key},
            json=self._get_json(session)
        )

        if r.status_code != 200:
            session.pop()
            raise RuntimeError(
                f"{r.status_code} {r.reason_phrase} {r.text}",
            )

        return FullResponse.model_validate_json(r.text)

    async def send(self,
                   prompt: Prompt,
                   role: Role = Role.User,
                   session: Session | None = None,
                   ) -> str:
        r = await self.send_raw(prompt, role, session)
        return cast(Message.TextSegment,
                    r.candidates[0].content.content[0]).text


class Session(BaseModel, Sequence[Message]):
    bot: Bot | None = Field(exclude=True)
    messages: list[Message] = Field(default_factory=list)
    message_lock: int = 0
    """Prevent earliest N messages from being trimmed"""

    async def trim(self, target_max: int | None = None, bot: Bot | None = None) -> int:
        """
        Trim the session until it's less than `target_max` tokens.
        Returns the total number of tokens after trimming.
        """
        bot = self.bot if bot is None else bot
        if bot is None:
            raise ValueError("bot is not set")
        if target_max is None:
            model_info = await bot.get_model_info()
            model_max_tokens = model_info.input_limit
            if bot.comp_tokens < 1:
                target_max = int((1-bot.comp_tokens) * model_max_tokens)
            else:
                target_max = max(0, model_max_tokens-int(bot.comp_tokens))

        message_lock = max(0, self.message_lock)
        msg_cnt = 0
        num_tokens = 0
        for message in self.messages[:message_lock]:
            num_tokens += await bot.count_tokens(message)
        for message in reversed(self.messages[message_lock:]):
            this_tokens = await bot.count_tokens(message)
            if num_tokens + this_tokens >= target_max:
                break
            msg_cnt += 1
            num_tokens += this_tokens
        trim_start = len(self.messages) - msg_cnt
        trim_start += trim_start % 2
        self.messages = self.messages[:message_lock] + self.messages[trim_start:]
        return num_tokens

    @validate_call
    def append(self, msg: Message) -> None:
        self.messages.append(msg)

    def pop(self, index: int = -1) -> Message:
        return self.messages.pop(index)

    def rollback(self, num: int):
        """
        Roll back `num` messages.
        """
        self.messages = self.messages[:len(self.messages)-num]

    def clear(self):
        """
        Clear the session.
        """
        self.messages.clear()

    @validate_call
    async def stream(self,
                     prompt: Prompt,
                     role: Role = Role.User,
                     ) -> AsyncGenerator[str, None]:
        """
        Stream messages from the bot.

        ## Parameters
        - `prompt` (str): What to say
        - `role` (Role): Role of the speaker
        """
        content = ''
        if self.bot is None:
            raise ValueError("bot is not set")
        async for r in self.bot.stream(prompt, role, self):
            content += r
            yield r

        self.append(Message(role=Role.Model, content=content))

    @validate_call
    async def send(self,
                   prompt: Prompt,
                   role: Role = Role.User,
                   ) -> str:
        """
        Send a message to the bot.

        ## Parameters
        - `prompt` (str): What to say
        - `role` (Role): Role of the speaker
        """
        if self.bot is None:
            raise ValueError("bot is not set")
        ret = await self.bot.send(prompt, role, self)
        self.append(Message(role=Role.Model, content=ret))
        return ret

    def __iter__(self):
        return iter(self.messages)

    def __len__(self):
        return len(self.messages)

    def __bool__(self):
        return bool(self.messages)

    def __reversed__(self):
        return reversed(self.messages)

    @overload
    def __getitem__(self, index: int) -> Message:
        ...

    @overload
    def __getitem__(self, s: slice) -> list[Message]:
        ...

    def __getitem__(self, index: int | slice):
        return self.messages[index]

    @validate_call
    def __setitem__(self, index: int, value: Message):
        self.messages[index] = value



import argparse
import asyncio
import ujson as json
import urllib.request
import re
from traceback import format_exc
from pathlib import Path
from getpass import getpass
from gemnine import bot, __author__, __version__
from platformdirs import user_cache_dir, user_config_dir


CONFIG_VERSION = "1.0.0"
SESSION_VERSION = "1.0.0"


async def main():
    parser = argparse.ArgumentParser(description="Gemnine")

    cache_dir = Path(user_cache_dir(
        "gemnine", __author__, SESSION_VERSION, ensure_exists=True))
    config_dir = Path(user_config_dir(
        "gemnine", __author__, CONFIG_VERSION, ensure_exists=True))

    parser.add_argument("-m", "--model", help="Model name", default=None)
    parser.add_argument("-c", "--config", help="Path to config file",
                        default=config_dir / "config.json")
    parser.add_argument("-s", "--session",
                        help="Path to session file", default=cache_dir / "session.json")
    parser.add_argument("-V", "--version", action="version",
                        version=f"%(prog)s {__version__}")

    args = parser.parse_args()

    model = args.model
    config_path = Path(args.config)
    session_path = Path(args.session)

    if not config_path.exists():
        print(f"Config file not found: {config_path}")
        user_input = input("Create a new config file? (Y/n): ").strip() or 'y'
        create_config = user_input.lower() == 'y'

        api_key = getpass("Enter your API key: ")
        sys_proxies = urllib.request.getproxies()
        proxy = sys_proxies.get("https", sys_proxies.get("http", None))
        b = bot.Bot(
            model=model or "models/gemini-pro",
            api_key=api_key, timeout=20, proxy=proxy)
        print(f"Using system proxy: {b.proxy}")

        if not model:
            print("Choose a model from the following list:")
            models = await b.list_models()
            for i, m in enumerate(models):
                print(f"    {i:2}: {m.name} - {m.description}")
            try:
                model = models[int(input("Enter the number of the model: "))]
                print(f"Selected model: {model.name}")
            except (ValueError, IndexError):
                print("Invalid input, exiting")
                exit(1)
        if create_config:
            config_path.write_text(b.model_dump_json(indent=4))
            print(f"Config file created: {config_path}")

    else:
        print(f"Loading config file: {config_path}")
        b = bot.Bot.model_validate_json(config_path.read_text())

        if model:
            b.model = model

    if not session_path.exists():
        print(f"Session file not found: {session_path}")
        print("Starting new session")
        session = b.new_session()
    else:
        print(f"Loading session file: {session_path}")
        with open(session_path, 'r') as f:
            session = bot.Session(bot=b, **json.load(f))

    if not b.api_key:
        b.api_key = getpass("Enter your API key: ")

    if b.proxy is None:
        sys_proxies = urllib.request.getproxies()
        proxy = sys_proxies.get("https", sys_proxies.get("http", None))
        b.proxy = proxy
        print(f"Using system proxy: {b.proxy}")

    print("\nChat started:", end="\n\n")
    namemap = {
        "user": "You",
        "model": "Bot",
        "function": "Function",
    }
    revisit_n = 2
    if len(session) > revisit_n + 1:
        print("...", end="\n\n")
    for h in session[-revisit_n:]:
        print(f"{namemap.get(h.role, h.role)}: {str(h)}", end="\n\n")

    try:
        while True:
            try:
                image_re = re.compile(r"(?<!#)(?:##)*#{1}image\((.*?)\)")
                role = bot.Role.User
                user_input = input("You: ")
                print('')

                if user_input.startswith('/'):
                    match user_input[1:].split():
                        case ["exit"]:
                            break
                        case ["save"]:
                            session_path.write_text(session.model_dump_json(indent=4))
                            print(f"Session saved to {session_path}")
                        case ["save", path]:
                            session_path = Path(path)
                            session_path.write_text(session.model_dump_json(indent=4))
                            print(f"Session saved to {session_path}")
                        case ["load"]:
                            with open(session_path, 'r') as f:
                                session = bot.Session(bot=b, **json.load(f))
                            print(f"Session loaded from {session_path}")
                        case ["load", path]:
                            session_path = Path(path)
                            with open(session_path, 'r') as f:
                                session = bot.Session(bot=b, **json.load(f))
                            print(f"Session loaded from {session_path}")
                        case ["clear"]:
                            session.clear()
                            print("Session cleared")
                        case ["rollback"]:
                            session.rollback(1)
                            print("Session rolled back")
                        case ["rollback", num]:
                            session.rollback(int(num))
                            print(f"Session rolled back {num} steps")
                        case ["role"]:
                            print(f"Current role {role.value}")
                            print("Choose a role from the following list:")
                            roles = tuple(bot.Role)
                            for i, r in enumerate(roles):
                                print(f"    {i:2}: {r.value}")
                            try:
                                role = roles[int(
                                    input("Enter the number of the role: "))]
                                print(f"Selected role: {role.value}")
                            except (ValueError, IndexError):
                                print("Invalid input")
                        case ["model"]:
                            print(f"Current model {b.model}")
                            print("Choose a model from the following list:")
                            models = await b.list_models()
                            for i, m in enumerate(models):
                                print(f"    {i:2}: {m.name}")
                            try:
                                model = models[int(
                                    input("Enter the number of the model: "))]
                                print(f"Selected model: {model.name}")
                                b.model = model.name
                            except (ValueError, IndexError):
                                print("Invalid input")
                        case ["tokens"]:
                            print(f"Used tokens(approximately): {await session.trim()}")
                        case ["help"]:
                            print("Commands:")
                            print("\t/exit")
                            print("\t/save")
                            print("\t/save <path>")
                            print("\t/load")
                            print("\t/load <path>")
                            print("\t/clear")
                            print("\t/rollback")
                            print("\t/rollback <num>")
                            print("\t/role")
                            print("\t/model")
                            print("\t/tokens")
                            print("\t/help")
                            print("\tTo embed an image, use #image(<url>)")
                        case _:
                            print("Invalid command, use /help for a list of commands.")
                    print('')
                    continue

                prompt: bot.Prompt
                if match := image_re.search(user_input):
                    prompt = list[bot.SegTypes]()
                    try:
                        for img_url in match.groups():
                            text_end = user_input.index(img_url)
                            pre_text = user_input[:text_end-7]  # remove `#image(`
                            if pre_text:
                                prompt.append(bot.Message.TextSegment(text=pre_text))
                            user_input = user_input[text_end +
                                                    len(img_url) + 1:]  # remove `)`
                            img_seg = bot.Message.ImageSegment(image_url=img_url)
                            prompt.append(img_seg)
                            if user_input:
                                prompt.append(bot.Message.TextSegment(text=user_input))
                    except ValueError as e:
                        print("Invalid image URL")
                        print(f"Error: {e}")
                        print(format_exc(), end="\n\n")
                        continue
                else:
                    prompt = user_input

                print("Bot: ", end='', flush=True)
                try:
                    async for res in session.stream(prompt, role):
                        print(res, end='', flush=True)
                    # print(await session.send(prompt=prompt, role=role))
                except (KeyboardInterrupt, asyncio.CancelledError):
                    print("\n<Interrupted>")
                finally:
                    print('\n')
            except KeyboardInterrupt:
                print("Exiting")
                break
            except Exception as e:
                print(f"Error: {e}")
                print(format_exc(), end="\n\n")
                continue
    finally:
        user_input = input("Save session? (Y/n): ")
        user_input = user_input.strip() or 'y'
        if user_input.lower() == 'y':
            session_path.write_text(session.model_dump_json(indent=4))
            print(f"Session saved to {session_path}")

asyncio.run(main())
